
--[[

   Copyright (c) 2016 Idiap Research Institute, http://www.idiap.ch/
   Written by Francois Fleuret <francois.fleuret@idiap.ch>

   This file is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   It is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
   License for more details.

   You should have received a copy of the GNU General Public License
   along with this file.  If not, see <http://www.gnu.org/licenses/>.

]]--

require 'torch'
require 'nn'

local DAG, parent = torch.class('nn.DAG', 'nn.Container')

function DAG:__init()
   parent.__init(self)
   -- Nodes are indexed by the module they contain
   self.node = { }
end

function DAG:createNode(nnm)
   if not self.node[nnm] then
      self:add(nnm) -- Add it to the object as a Container
      local node = {}
      node.succ = {}
      node.pred = {}
      node.index = #self.modules
      self.node[nnm] = node
   end
end

-- The main use should be to add an edge between two modules, but it
-- can also add a full sequence of modules
function DAG:addEdge(...)
   self.sorted = nil
   local prev
   for _, nnm in pairs({...}) do
      self:createNode(nnm)
      if prev then
         table.insert(self.node[nnm].pred, prev)
         table.insert(self.node[prev].succ, nnm)
      end
      prev = nnm
   end
end

-- Apply f on t recursively; use the corresponding element from args
-- (i.e. same keys) as second parameter to f when available; return
-- the results from f, organized in a similarly nested table.
function DAG:nestedApply(f, t, args)
   if torch.type(t) == 'table' then
      local result = {}
      for k, s in pairs(t) do
         result[k] = self:nestedApply(f, s, args and args[k])
      end
      return result
   else
      return f(t, args)
   end
end

function DAG:setInput(i)
   self.sorted = nil
   self.inputModules = i
   self:nestedApply(
      function(nnm)
         if #self.node[nnm].succ == 0 then
            error('Input modules must have outgoing  edges.')
         end
         if #self.node[nnm].pred > 0 then
            error('Input modules cannog have incoming edges.')
         end
      end,
      self.inputModules
   )
end

function DAG:setOutput(o)
   self.sorted = nil
   self.outputModules = o
   self:nestedApply(
      function(nnm)
         if #self.node[nnm].pred == 0 then
            error('Output module must have incoming edges.')
         end
         if #self.node[nnm].succ > 0 then
            error('Output module cannot have outgoing edges.')
         end
      end,
      self.outputModules
   )
end

function DAG:putInOrder()
   if self.sorted then
      return
   end

   local distance = {}
   self:nestedApply(function(m) distance[m] = 1 end, self.inputModules)

   local nc
   repeat
      nc = 0
      for nnma, node in pairs(self.node) do
         for _, nnmb in pairs(node.succ) do
            if distance[nnma] and (not distance[nnmb] or distance[nnmb] < distance[nnma] + 1) then
               distance[nnmb] = distance[nnma] + 1
               nc = nc + 1
            end
         end
      end
   until nc == 0

   self.sorted = { }
   for m, d in pairs(distance) do
      table.insert(self.sorted, { distance = d, nnm = m })
   end

   table.sort(self.sorted, function(a, b) return a.distance < b.distance end)

   for i, a in ipairs(self.sorted) do self.sorted[i] = a.nnm end
end

function DAG:computeGradOutput(gradInputSucc)
   local gi
   if #gradInputSucc == 1 then
      gi = gradInputSucc[1] -- we avoid a clone()
   elseif #gradInputSucc > 1 then
      for k = 1, #gradInputSucc do
         if gi then
            gi:add(gradInputSucc[k])
         else
            gi = gradInputSucc[k]:clone()
         end
      end
   end
   return gi
end

function DAG:print()
   self:putInOrder()

   for i, d in ipairs(self.sorted) do
      print('#' .. i .. ' -> ' .. torch.type(d))
   end
end

----------------------------------------------------------------------

function DAG:updateOutput(input)
   self:putInOrder()

   self:nestedApply(
      function(nnm, i)
         self.node[nnm].input = i
         -- nnm:updateOutput(i)
         self:rethrowErrors(nnm, self.node[nnm].index, 'updateOutput', i)
      end,
      self.inputModules,
      input
   )

   for _, nnm in ipairs(self.sorted) do
      local node = self.node[nnm]
      if #node.pred > 0 then
         local i
         if #node.pred == 1 then
            i = node.pred[1].output
         elseif #node.pred > 1 then
            i = {}
            for k = 1, #node.pred do
               i[k] = node.pred[k].output
            end
         end
         node.input = i
         -- nnm:updateOutput(i)
         self:rethrowErrors(nnm, self.node[nnm].index, 'updateOutput', i)
      end
   end

   self.output = self:nestedApply(
      function(m) return m.output end,
      self.outputModules
   )

   return self.output
end

function DAG:updateGradInput(input, gradOutput)
   assert(self.sorted, 'there has been a DAG structure change before a DAG:updateGradInput')

   self:nestedApply(
      function(nnm, go)
         -- nnm:updateGradInput(self.node[nnm].input, go)
         self:rethrowErrors(nnm, self.node[nnm].index, 'updateGradInput', self.node[nnm].input, go)
      end,
      self.outputModules, gradOutput
   )

   self:nestedApply(
      function(nnm, i) self.node[nnm].input = i end,
      self.inputModules, input
   )

   for _, node in pairs(self.node) do
      node.gradInputSucc = {}
   end

   for k = #self.sorted, 1, -1 do
      local nnm = self.sorted[k]
      local node = self.node[nnm]
      local pred, gradInputSucc = node.pred, node.gradInputSucc

      if #gradInputSucc > 0 then
         node.gradOutput = self:computeGradOutput(gradInputSucc)
         -- nnm:updateGradInput(node.input, node.gradOutput)
         self:rethrowErrors(nnm, self.node[nnm].index, 'updateGradInput', node.input, node.gradOutput)
      end

      -- We fill the gradInputSucc of our predecessors
      if #pred == 1 then
         table.insert(self.node[pred[1]].gradInputSucc, nnm.gradInput)
      elseif #pred > 1 then
         if not torch.type(nnm.gradInput) == 'table' then
            error('Should have a table gradInput since it has multiple predecessors')
         end
         for n = 1, #pred do
            table.insert(self.node[node.pred[n]].gradInputSucc, nnm.gradInput[n])
         end
      end
   end

   self.gradInput = self:nestedApply(function(m) return m.gradInput end, self.inputModules)

   return self.gradInput
end

function DAG:accGradParameters(input, gradOutput, scale)
   scale = scale or 1

   assert(self.sorted, 'there has been a DAG structure change before a DAG:accGradParameters')

   for k = 1, #self.modules do
      local nnm = self.modules[k]
      local node = self.node[nnm]
      -- nnm:accGradParameters(node.input, node.gradOutput, scale)
      self:rethrowErrors(nnm, k, 'accGradParameters', node.input, self:computeGradOutput(node.gradInputSucc), scale)
   end
end

----------------------------------------------------------------------
